<a class="nav-link dropdown-toggle" href="#" id="navbarDropdown" role="button" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">
  Version {{ version }}
</a>
<div class="dropdown-menu" aria-labelledby="navbarDropdown" id="version-menu"></div>
