#!/usr/bin/env python3
#
# compliance_checker documentation build configuration file, created by
# sphinx-quickstart on Mon Oct  9 21:28:42 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

from datetime import datetime

import compliance_checker

version = release = compliance_checker.__version__
cur_date = datetime.utcnow()


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = "2.1"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "numpydoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

# Make MyST generate anchors for headings
myst_heading_anchors = 2

autodoc_typehints = "description"

# Tweak how docs are formatted
napoleon_use_rtype = False

# Control main class documentation
autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Custom sidebar templates, maps document names to template names.
html_sidebars = {"**": ["versions", "sidebar-nav-bs"]}

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]

# Controlling automatically generating summary tables in the docs
autosummary_generate = True
autosummary_ignore_module_all = False

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "compliance_checker"
copyright = f"{cur_date:%Y}, Dave Foster"
author = "Dave Foster"


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "external_links": [
        {
            "name": "Release Notes",
            "url": "https://github.com/ioos/compliance-checker/releases",
        },
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/ioos/compliance-checker",
            "icon": "fab fa-github-square",
        },
    ],
    "use_edit_page_button": False,
    "navbar_align": "left",
    "navbar_start": ["navbar-logo"],
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["search-field", "navbar-icon-links"],
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "compliance_checkerdoc"


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, "compliance_checker", "compliance_checker Documentation", [author], 1),
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "compliance_checker",
        "compliance_checker Documentation",
        author,
        "compliance_checker",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# GitHub issuecomment anchors are returning as broken for some reason.
linkcheck_ignore = ["https://github.com/ioos/compliance-checker/issues/.*"]
